@@ -20,8 +20,6 @@
  * Author: Christian Kellner <gicmo@gnome.org> 
  */
 
-/* WARNING! MOST OF THIS CODE IS TOTALLY UNTESTED! ck - 20.06.2005 */
-
 #include <config.h>
 #include <string.h>
 #include <unistd.h>
@@ -542,7 +540,7 @@ xp_object_get_number (xmlXPathObjectPtr result)
 
 /*** *** *** *** *** *** */
 #define XPATH_HREF "string(/D:multistatus/D:response[%d]/D:href)"
-#define XPATH_STATUS "string(/D:multistatus/D:response[%d]/D:status)"
+#define XPATH_STATUS "string(/D:multistatus/D:response[%d]/D:propstat/D:status)"
 #define XPATH_GETETAG_STATUS "string(/D:multistatus/D:response[%d]/D:propstat/D:prop/D:getetag/../../D:status)"
 #define XPATH_GETETAG "string(/D:multistatus/D:response[%d]/D:propstat/D:prop/D:getetag)"
 #define XPATH_CALENDAR_DATA "string(/D:multistatus/D:response[%d]/C:calendar-data)"
@@ -829,6 +827,8 @@ caldav_server_list_objects (ECalBackendCalDAV *cbdav, CalDAVObject **objs, int *
 	message = soup_message_new ("REPORT", priv->uri);
 	soup_message_add_header (message->request_headers, 
 				 "User-Agent", "Evolution/" VERSION);
+	soup_message_add_header (message->request_headers, 
+				 "Depth", "1");
 
 	soup_message_set_request (message, 
 				  "text/xml",
@@ -1325,11 +1325,16 @@ initialize_backend (ECalBackendCalDAV *cbdav)
 		priv->need_auth = TRUE;
 	}
 	
+	os_val = e_source_get_property(source, "ssl");
 	uri = e_cal_backend_get_uri (E_CAL_BACKEND (cbdav));
 
 	if (g_str_has_prefix (uri, "caldav://")) {
-		priv->uri = g_strconcat ("http://", uri + 9, NULL);
-	} else {
+		if (os_val && os_val[0] == '1') {
+			priv->uri = g_strconcat ("https://", uri + 9, NULL);
+		} else {
+			priv->uri = g_strconcat ("http://", uri + 9, NULL);
+		}
+	} else {		
 		priv->uri = g_strdup (uri);
 	} 
 		
@@ -1449,7 +1454,39 @@ caldav_remove (ECalBackendSync *backend,
 	return GNOME_Evolution_Calendar_Success;
 }
 
-#define pack_cobj(__cobj) (g_strdup_printf ("BEGIN:VCALENDAR\n%sEND:VCALENDAR\n", __cobj))
+
+static char *
+pack_cobj (ECalBackendCalDAV *cbdav, ECalComponent *ecomp) 
+{
+	ECalBackendCalDAVPrivate *priv;
+	icalcomponent *calcomp;
+	icalcomponent *icomp;
+	char          *objstr;
+	
+	priv  = E_CAL_BACKEND_CALDAV_GET_PRIVATE (cbdav);
+
+	icomp = e_cal_component_get_icalcomponent (ecomp);
+
+	if (icalcomponent_isa (icomp) != ICAL_VCALENDAR_COMPONENT) {
+		icalcomponent *cclone;
+
+		calcomp = e_cal_util_new_top_level ();
+		cclone = icalcomponent_new_clone (icomp);
+		icalcomponent_add_component (calcomp, cclone);
+		e_cal_util_add_timezones_from_component(calcomp,
+							cclone);
+	} else {
+		calcomp = icalcomponent_new_clone (icomp);
+	}
+
+	objstr = icalcomponent_as_ical_string (calcomp);
+	
+	g_assert (objstr);
+	
+	return g_strdup (objstr);
+		
+}
+
 
 static ECalBackendSyncStatus
 caldav_create_object (ECalBackendSync  *backend, 
@@ -1490,7 +1527,7 @@ caldav_create_object (ECalBackendSync  *backend,
 	
 		object.href  = href;
 		object.etag  = NULL;
-		object.cdata = pack_cobj (*calobj);
+		object.cdata = pack_cobj (cbdav, comp);
 
 		status = caldav_server_put_object (cbdav, &object);
 
@@ -1567,7 +1604,7 @@ caldav_modify_object (ECalBackendSync  *backend,
 	
 		object.href  = g_strdup (e_cal_component_get_href (cache_comp));
 		object.etag  = g_strdup (e_cal_component_get_etag (cache_comp));
-		object.cdata = pack_cobj (calobj);
+		object.cdata = pack_cobj (cbdav, comp);
 
 		status = caldav_server_put_object (cbdav, &object);
 
@@ -1675,15 +1712,296 @@ caldav_discard_alarm (ECalBackendSync *backend,
 	return GNOME_Evolution_Calendar_Success;
 }
 
+/* FIXME: use list here? */
+static ECalBackendSyncStatus
+extract_objects (icalcomponent       *icomp,
+		 icalcomponent_kind   ekind,
+		 GList              **objects)
+{
+	icalcomponent         *scomp;
+	icalcomponent_kind     kind;
+	
+	g_return_val_if_fail (icomp, GNOME_Evolution_Calendar_OtherError);
+	g_return_val_if_fail (objects, GNOME_Evolution_Calendar_OtherError);
+	
+	kind = icalcomponent_isa (icomp);
+	
+	if (kind == ekind) {
+		*objects = g_list_prepend (NULL, icomp);
+		return GNOME_Evolution_Calendar_Success;
+	}
+
+	if (kind != ICAL_VCALENDAR_COMPONENT) {
+		return GNOME_Evolution_Calendar_InvalidObject;
+	}
+
+	*objects = NULL;
+	scomp = icalcomponent_get_first_component (icomp, 
+						   ekind);
+
+	while (scomp) {
+		
+		/* Remove components from toplevel here */
+		*objects = g_list_prepend (*objects, scomp);
+		icalcomponent_remove_component (icomp, scomp);
+		
+		scomp = icalcomponent_get_next_component (icomp, ekind);
+	}
+
+	return GNOME_Evolution_Calendar_Success;
+}
+
+#define is_error(__status) (__status != GNOME_Evolution_Calendar_Success)
+
+static ECalBackendSyncStatus
+process_object (ECalBackendCalDAV   *cbdav,
+		ECalComponent       *ecomp,
+		gboolean             online,
+		icalproperty_method  method)
+{
+	ECalBackendCalDAVPrivate *priv;
+	ECalBackendSyncStatus     status;
+	ECalBackend              *backend;
+	ECalComponent            *ccomp;
+	struct icaltimetype       now;
+	const char               *uid;	
+	const char               *rid;
+	char                     *ostr;
+	char                     *oostr;
+	
+	priv = E_CAL_BACKEND_CALDAV_GET_PRIVATE (cbdav);
+	backend = E_CAL_BACKEND (cbdav);
+	
+	/* ctime, mtime */
+	now = icaltime_from_timet (time (NULL), 0);
+	e_cal_component_set_created (ecomp, &now);
+	e_cal_component_set_last_modified (ecomp, &now);
+
+	e_cal_component_get_uid (ecomp, &uid);
+	rid = e_cal_component_get_recurid_as_string (ecomp);
+
+	ccomp = e_cal_backend_cache_get_component (priv->cache, uid, NULL);
+
+	if (ccomp != NULL) {
+		oostr = e_cal_component_get_as_string (ccomp);	
+	} else {
+		oostr = NULL;
+	}
+	
+	ostr = e_cal_component_get_as_string (ecomp);	
+	
+	status = GNOME_Evolution_Calendar_Success;
+	
+	switch (method) {
+	
+		case ICAL_METHOD_PUBLISH:
+		case ICAL_METHOD_REQUEST:
+		case ICAL_METHOD_REPLY:
+
+		if (online) {
+			CalDAVObject object = { 0, };
+
+			if (ccomp) {
+				const char *href;
+				const char *etag;
+
+				href = e_cal_component_get_href (ccomp);
+				etag = e_cal_component_get_etag (ccomp);
+				
+				object.href  = g_strdup (href);
+				object.etag  = g_strdup (href);
+
+			} else {
+				object.href = e_cal_component_gen_href (ecomp, priv->uri);
+			}
+			
+			object.cdata = pack_cobj (cbdav, ecomp);
+			status = caldav_server_put_object (cbdav, &object);
+			e_cal_component_set_etag (ecomp, object.etag);
+			caldav_object_free (&object, FALSE);
+		} else {
+			ECalComponentSyncState sstate;
+			
+			if (ccomp) {
+				sstate = E_CAL_COMPONENT_LOCALLY_MODIFIED;
+			} else {
+				sstate = E_CAL_COMPONENT_LOCALLY_CREATED;
+			}
+
+			e_cal_component_set_synch_state (ecomp, sstate); 
+
+		}
+		
+		if (status != GNOME_Evolution_Calendar_Success) {
+			break;
+		}
+
+		e_cal_backend_cache_put_component (priv->cache, ecomp);
+			
+		if (ccomp) {
+
+			e_cal_backend_notify_object_modified (backend,
+							      ostr,
+							      oostr);
+
+		} else {
+				
+			e_cal_backend_notify_object_created (backend,
+							     ostr);
+		}
+		
+		break;
+
+
+		case ICAL_METHOD_CANCEL:
+	
+			if (ccomp == NULL) {
+				status = GNOME_Evolution_Calendar_ObjectNotFound;
+				break;
+			}
+	
+			/* FIXME: this is not working for instances
+			 * of recurring appointments - yet - */	
+			if (online) {
+				CalDAVObject object;
+				const char *href;
+				const char *etag;
+
+				href = e_cal_component_get_href (ccomp);
+				etag = e_cal_component_get_etag (ccomp);
+				
+				object.href  = g_strdup (href);
+				object.etag  = g_strdup (href);
+				object.cdata = NULL;
+
+				status = caldav_server_delete_object (cbdav,
+								      &object);
+			
+				caldav_object_free (&object, FALSE);
+				
+			} else {
+				/* mark component as out of synch */
+				e_cal_component_set_synch_state (ecomp, 
+				E_CAL_COMPONENT_LOCALLY_DELETED);
+				
+			}	
+		
+			if (status != GNOME_Evolution_Calendar_Success) {
+				break;
+			}
+
+			e_cal_backend_cache_remove_component (priv->cache,
+							      uid,
+							      rid);
+			
+			e_cal_backend_notify_object_removed (E_CAL_BACKEND (backend),
+							     uid,
+							     oostr,
+							     ostr);
+			break;
+
+		default:
+			/* WTF ? */
+			status = GNOME_Evolution_Calendar_UnsupportedMethod;
+			break;
+	}
+
+	g_object_unref (ecomp);
+	g_free (ostr);
+	g_free (oostr);
+	
+	if (ccomp) {
+		g_object_unref (ccomp);
+	}
+
+	return status;
+}
 
 static ECalBackendSyncStatus
 caldav_receive_objects (ECalBackendSync *backend,
 			EDataCal        *cal, 
 			const char      *calobj)
 {
-	/* FIXME: implement me! */
-	g_warning ("function not implemented %s", G_STRFUNC);
-	return GNOME_Evolution_Calendar_OtherError;	
+	ECalBackendCalDAV        *cbdav;
+	ECalBackendCalDAVPrivate *priv;
+	ECalBackendSyncStatus     status;
+	icalcomponent            *icomp;
+	icalcomponent_kind        kind;
+	icalproperty_method       tmethod;
+	gboolean                  online;
+	GList                    *timezones;
+	GList                    *objects;
+	GList                    *iter;
+	
+	cbdav = E_CAL_BACKEND_CALDAV (backend);
+	priv  = E_CAL_BACKEND_CALDAV_GET_PRIVATE (cbdav);
+	
+	icomp = icalparser_parse_string (calobj);
+	
+	/* Try to parse cal object string */
+	if (icomp == NULL) {
+		return GNOME_Evolution_Calendar_InvalidObject;
+	}
+
+	/* FIXME: use the e_cal_backend_xxx_kind call here */
+	kind = ICAL_VEVENT_COMPONENT;
+	status = extract_objects (icomp, kind, &objects);
+
+	if (status != GNOME_Evolution_Calendar_Success) {
+		return status;
+	}
+
+	/* Extract optional timezone compnents */
+	kind = ICAL_VTIMEZONE_COMPONENT;
+	status = extract_objects (icomp, kind, &timezones);
+	
+	if (status == GNOME_Evolution_Calendar_Success) {
+		/* FIXME: */
+		/* Do something usefull with the timezone */
+	}
+	
+	/*   */
+	g_mutex_lock (priv->lock);
+
+	status = check_state (cbdav, &online);
+
+	if (status != GNOME_Evolution_Calendar_Success) {
+		/* FIXME: free components here */
+		g_mutex_unlock (priv->lock);
+		return status;
+	}
+	
+	tmethod = icalcomponent_get_method (icomp);
+	
+	for (iter = objects; iter && ! is_error (status); iter = iter->next) {
+		icalcomponent       *scomp;
+		ECalComponent       *ecomp;
+		icalproperty_method  method;
+		
+		scomp = (icalcomponent *) iter->data;
+		ecomp = e_cal_component_new ();
+
+		e_cal_component_set_icalcomponent (ecomp, scomp);
+
+		if (icalcomponent_get_first_property (scomp,
+						      ICAL_METHOD_PROPERTY)) {
+			
+			method = icalcomponent_get_method (scomp);
+		} else {
+			method = tmethod;
+		}
+		
+		status = process_object (cbdav, ecomp, online, method);		
+		
+		g_object_unref (ecomp);
+	}
+
+	g_list_free (objects);
+	g_list_free (timezones);
+	
+	g_mutex_unlock (priv->lock);
+
+	return status;	
 }
 
 static ECalBackendSyncStatus
@@ -1692,10 +2010,11 @@ caldav_send_objects (ECalBackendSync  *backend,
 		     const char       *calobj, 
 		     GList           **users,
 		     char            **modified_calobj)
-{
-	/* FIXME: implement me! */
-	g_warning ("function not implemented %s", G_STRFUNC);
-	return GNOME_Evolution_Calendar_OtherError;	
+{	
+	*users = NULL;
+	*modified_calobj = g_strdup (calobj);
+
+	return GNOME_Evolution_Calendar_Success;
 }
 
 static ECalBackendSyncStatus
@@ -1750,8 +2069,8 @@ caldav_get_timezone (ECalBackendSync  *backend,
 {
 	ECalBackendCalDAV        *cbdav;
 	ECalBackendCalDAVPrivate *priv;
-	icaltimezone *zone;
-	icalcomponent *icalcomp;
+	const icaltimezone       *zone;
+	icalcomponent            *icalcomp;
 
 	cbdav = E_CAL_BACKEND_CALDAV (backend);
 	priv  = E_CAL_BACKEND_CALDAV_GET_PRIVATE (cbdav);
@@ -1770,10 +2089,12 @@ caldav_get_timezone (ECalBackendSync  *backend,
 		}
 	}
 
-	icalcomp = icaltimezone_get_component (zone);
-	if (!icalcomp)
+	icalcomp = icaltimezone_get_component ((icaltimezone *) zone);
+	
+	if (!icalcomp) {
 		return GNOME_Evolution_Calendar_InvalidObject;
-
+	}
+		
 	*object = g_strdup (icalcomponent_as_ical_string (icalcomp));
 
 	return GNOME_Evolution_Calendar_Success;
